@@ -209,49 +209,36 @@ def atomic_add(API):
     
     ocl_add = """
     // #pragma OPENCL EXTENSION cl_khr_int64_base_atomics : enable
-    KERNEL void atomic_add_float( 
-            GLOBAL_MEM float *ptr, 
-            const float temp) 
-    {
-    // The work-around of AtomicAdd for float
-    // lockless add *source += operand 
-    // Caution!!!!!!! Use with care! You have been warned!
-    // http://simpleopencl.blogspot.com/2013/05/atomic-operations-and-floats-in-opencl.html
-    // Source: https://github.com/clMathLibraries/clSPARSE/blob/master/src/library/kernels/csrmv_adaptive.cl
-        union {
-            unsigned int intVal;
-            float floatVal;
-        } newVal;
-        
-        union {
-            unsigned int intVal;
-            float floatVal;
-        } prevVal;
-        
-        do {
-            prevVal.floatVal = *ptr;
-            newVal.floatVal = prevVal.floatVal + temp;
-        } while (atomic_cmpxchg((__global volatile unsigned int *)ptr, prevVal.intVal, newVal.intVal) != prevVal.intVal);
-    };  
-    
-    KERNEL void atomic_add_float2_old( 
+ 
+    KERNEL void atomic_add_float2_typen( 
             GLOBAL_MEM float2 *ptr, 
             const float2 temp) 
-    {
-    atomic_add_float( 
-            (GLOBAL_MEM float*)ptr, 
-            temp.x);
-    atomic_add_float( 
-    (GLOBAL_MEM float*)ptr+1, 
-    temp.y);
-    };  
-    
+        { // Add atomic two sum for real/imaginary parts
+        
+        __global const float * realptr = ( __global float *)ptr;
+        __global const float * imagptr = ( __global float *)ptr + 1;
+        float prevVal;
+        float newVal;
+        do { // atomic add of the value
+            prevVal = *realptr ;
+            newVal = prevVal + temp.x;
+        } while (atomic_cmpxchg((__global volatile unsigned int *)realptr, as_int(prevVal), as_int(newVal)) != as_int(prevVal));
+ 
+        // End of real part
+        
+        // Now do the imaginary part
+        
+        do { // atomic add of the value
+            prevVal = *imaglptr ;
+            newVal = prevVal + temp.y;
+        } while (atomic_cmpxchg((__global volatile unsigned int *)imagptr, as_int(prevVal), as_int(newVal)) != as_int(prevVal));
+ 
+    };   
+ 
     KERNEL void atomic_add_float2( 
             GLOBAL_MEM float2 *ptr, 
-            const float2 temp, 
-            GLOBAL_MEM float2 *res) 
+            const float2 temp) 
         { // Add atomic two sum for real/imaginary parts
-          // *res array stores the error
         union {
             unsigned int intVal;
             float floatVal;
@@ -261,38 +248,16 @@ def atomic_add(API):
             unsigned int intVal;
             float floatVal;
         } prevVal;
-        float x;
-        float val;
-        float newacc;
-        float r;
+ 
             
         __global const float * realptr = ( __global float *)ptr;
         __global const float * imagptr = ( __global float *)ptr + 1;
-        __global const float * realres = ( __global float *)res;
-        __global const float * imagres = ( __global float *)res + 1;  
           
         do { // atomic add of the value
             prevVal.floatVal = *realptr ;
             newVal.floatVal = prevVal.floatVal + temp.x;
         } while (atomic_cmpxchg((__global volatile unsigned int *)realptr, prevVal.intVal, newVal.intVal) != prevVal.intVal);
-        
-        x = prevVal.floatVal;
-        val = temp.x;
-        
-        if (fabs(x) < fabs(val))
-        {
-            const float swap = x;
-            x = val;
-            val = swap;
-        }
-        newacc = x + val;
-        r = val - (newacc - x);    // Recover rounding error using Fast2Sum, assumes |oldacc|>=|val|
-                
-        do { // atomic_add(realres, r)
-            prevVal.floatVal = *realres ;
-            newVal.floatVal = prevVal.floatVal + r;
-        } while (atomic_cmpxchg((__global volatile unsigned int *)realres, prevVal.intVal, newVal.intVal) != prevVal.intVal);
-        
+ 
         // End of real part
         
         // Now do the imaginary part
@@ -301,69 +266,22 @@ def atomic_add(API):
             prevVal.floatVal = *imagptr;
             newVal.floatVal = prevVal.floatVal + temp.y;
         } while (atomic_cmpxchg((__global volatile unsigned int *)imagptr, prevVal.intVal, newVal.intVal) != prevVal.intVal);
-        
-                x = prevVal.floatVal; // this copy the old value to x
-                val = temp.y; // this copy the  real part to val
-                
-                if (fabs(x) < fabs(val)) 
-                {
-                    const float swap = x;
-                    x = val;
-                    val = swap;
-                }
-                newacc = x + val;
-                r = val - (newacc - x);    // Recover rounding error using Fast2Sum, assumes |oldacc|>=|val|
-                
-                do {
-                    prevVal.floatVal = *imagres ;
-                    newVal.floatVal = prevVal.floatVal + r;
-                } while (atomic_cmpxchg((__global volatile unsigned int *)imagres, prevVal.intVal, newVal.intVal) != prevVal.intVal);
-       
+   
     };  
     
+    
     """
     
     cuda_add = """
     
-    __device__ void atomic_add_float_twosum(
-            GLOBAL_MEM float * address,
-            float val,
-            GLOBAL_MEM float * residue)
-    {
-        
-        // https://devtalk.nvidia.com/default/topic/817899/atomicadd-kahan-summation/
-        // By Sylvain Collange
-        // Also add the branch clsparse
-        
-        float x = atomicAdd(address, val);
-        
-        if (fabs(x) < fabs(val))
-        {
-            const float swap = x;
-            x = val;
-            val = swap;
-        }
-        float newacc = x + val;
-        float r = val - (newacc - x);    // Recover rounding error using Fast2Sum, assumes |oldacc|>=|val|
-        atomicAdd(residue, r);    // Accumulate error in residue
-        
-    }    
-    
-    __device__ void atomic_add_float( 
-            GLOBAL_MEM float *ptr, 
-            const float temp) 
-    { // Wrapper around CUDA atomicAdd();
-    atomicAdd(ptr, temp); 
-    };   
     __device__ void atomic_add_float2( 
             GLOBAL_MEM float2 * ptr, 
-            const float2 temp,
-            GLOBAL_MEM float2 *res) 
+            const float2 temp) 
     { // Wrapper around CUDA atomicAdd();
-    // atomicAdd((float*)ptr, temp.x);
-    // atomicAdd((float*)ptr+1, temp.y);
-    atomic_add_float_twosum((float*)ptr, temp.x, (float*)res);
-    atomic_add_float_twosum((float*)ptr+1, temp.y, (float*)res+1);
+    atomicAdd((float*)ptr, temp.x);
+    atomicAdd((float*)ptr+1, temp.y);
+    // atomic_add_float_twosum((float*)ptr, temp.x, (float*)res);
+    // atomic_add_float_twosum((float*)ptr+1, temp.y, (float*)res+1);
     };       
     
     """  
@@ -605,7 +523,8 @@ def cSpmvh():
     """
     
     R="""
-        KERNEL void pELL_spmvh_mCoil_old(
+    
+        KERNEL void pELL_spmvh_mCoil(
         const    unsigned int    Reps,             // number of coils
         const    unsigned int    nRow,        // number of rows
         const    unsigned int    prodJd,     // product of Jd
@@ -616,31 +535,33 @@ def cSpmvh():
         GLOBAL_MEM const unsigned int *meshindex,    // meshindex, prodJd * dim
         GLOBAL_MEM const unsigned int *kindx,    // unmixed column indexes of all dimensions
         GLOBAL_MEM const float2 *udata,    // interpolation data before Kronecker product
-        GLOBAL_MEM    float    *kx, 
-        GLOBAL_MEM    float    *ky,
+        GLOBAL_MEM    float2    *k, 
+        //GLOBAL_MEM float2 *res,
         GLOBAL_MEM const float2 *input)   // y
         {      
         const unsigned int t = get_local_id(0);
-        const unsigned int d = get_local_size(0);
-        const unsigned int g = get_group_id(0);
-        const unsigned int a = g/prodJd;
-        const unsigned int b = g - (prodJd*a);
-        const unsigned int myRow = a*d + t;
-        const unsigned int j = b;
+        const unsigned int vecWidth=${LL};
+        // Thread ID within wavefront
+        const unsigned int id = t & (vecWidth-1);
+        
+        // One row per wavefront
+        unsigned int vecsPerBlock=get_local_size(0)/vecWidth;
+        unsigned int myRow=(get_group_id(0)*vecsPerBlock) + (t/ vecWidth); // the myRow-th non-Cartesian sample
         unsigned int m = myRow / Reps;
         unsigned int nc = myRow - m * Reps;
         
-        
-        // unsigned int myRow= get_global_id(0);
         float2 zero;
         zero.x = 0.0;
         zero.y = 0.0;
-        if (m< nRow)
-                {if (nc < Reps)
-                {
-                float2 u = zero;
-                //  for (unsigned int j = 0;  j  <  prodJd; j ++)
-                if (j<prodJd)
+        
+        
+        if (myRow < nRow * Reps)
+        {
+        const unsigned int vecStart = 0; 
+        const unsigned int vecEnd =prodJd;             
+        float2  u=zero;
+        
+        for (unsigned int j = vecStart+id;  j<vecEnd; j += vecWidth)
                         {    
                         // now doing the first dimension
                         unsigned int index_shift = m * sumJd;
@@ -666,17 +587,16 @@ def cSpmvh():
                         u.x =  spdata.x*ydata.x + spdata.y*ydata.y;
                         u.y =  - spdata.y*ydata.x + spdata.x*ydata.y;
                         
-                        atomic_add_float(kx + col*Reps + nc, u.x);
-                        atomic_add_float(ky + col*Reps + nc, u.y);
-                        
+                        atomic_add_float2(k + col*Reps + nc, u);//, res + col*Reps + nc);
+                        LOCAL_BARRIER;
+                        // atomic_add_float2(k + col*Reps + nc, u, res + col*Reps + nc);
                         }; // Iterate for (unsigned int j = 0;  j  <  prodJd; j ++)
-                }; // if (nc < Reps)
         };  // if (m < nRow)
         
-        };    // End of pELL_spmvh_mCoil_old  
+        };    // End of xELL_spmvh_mCoil    
+    
         
-        
-        KERNEL void pELL_spmvh_mCoil(
+        KERNEL void pELL_spmvh_mCoil_new(
         const    unsigned int    Reps,             // number of coils
         const    unsigned int    nRow,        // number of rows
         const    unsigned int    prodJd,     // product of Jd
@@ -723,7 +643,7 @@ def cSpmvh():
                     float2 ydata=input[myRow*Reps + nc]; // kout[col];
                     u.x =  spdata.x*ydata.x + spdata.y*ydata.y;
                     u.y =  - spdata.y*ydata.x + spdata.x*ydata.y;
-                    atomic_add_float2(k + col*Reps + nc, u, res + col*Reps + nc);
+                    atomic_add_float2(k + col*Reps + nc, u);
                         
                 }; // Iterate for (unsigned int j = 0;  j  <  prodJd; j ++)
                 
